#include <unistd.h>

#include <SDL_video.h>

#ifdef USE_LITMUS
#include "litmus.h"
#endif

#include "inc/common.h"
#include "inc/video_common.h"
#include "inc/options.h"
#include "inc/queue.h"
#include "inc/video_queue.h"
#include "inc/timing.h"
#include "inc/profile.h"
#include "inc/thread_ctx.h"
#include "inc/sem_common.h"
#include "inc/litmus_video.h"
#include "inc/display_common.h"

static SDL_Surface *screen = NULL;
static struct vid_q_item src_vq_item;
static AVPicture screen_pict;
static unsigned long long disp_frame_num = 0;
static struct {
	/* When we expect this frame to be put on the screen */
	struct timespec expected;
	/* When we put the last frame on the screen */
	struct timespec last;
	/* release time */
	struct timespec rel;
} tspecs = {
	.expected	= { .tv_sec = 0, .tv_nsec = 0 },
	.last		= { .tv_sec = 0, .tv_nsec = 0 },
	.rel		= { .tv_sec = 0, .tv_nsec = 0 },
};

/*
 * Must be called after initialize_av() so that the video information (like
 * width and height) is known.
 */
void initialize_display(struct thread_ctx *t_ctx)
{
	struct video_state *vs = t_ctx->vs;
	struct options *opts = t_ctx->opts;
	const int width = vs->codec_ctx->width;
	const int height = vs->codec_ctx->height;

	if (opts->use_sdl) {
		screen = SDL_SetVideoMode(width, height, 0, SDL_ASYNCBLIT);
		if (!screen)
			die("could not set video mode\n");

		/* hook up pict so copying draws into the screen */
		screen_pict.data[0] = screen->pixels;
		screen_pict.linesize[0] = screen->pitch;
	} else {
		/* no SDL, make a fake screen */
		avpicture_alloc_wrap(&screen_pict, width, height);
	}

	avpicture_alloc_wrap(&src_vq_item.picture, width, height);
}

static inline void update_timespecs(struct thread_ctx *t_ctx)
{
	static struct timespec period_spec = { .tv_sec = 0, .tv_nsec = 0 };
	long long resp_time;
	unsigned long long t_ns;
	mono_time_t t_mono;

	if (tspec_uninit(&period_spec))
	{
		period_spec.tv_sec  = t_ctx->period / NSEC_PER_SEC;
		period_spec.tv_nsec = t_ctx->period -
			((unsigned long long)period_spec.tv_sec) * NSEC_PER_SEC;
	}

	/* don't increment release time yet */
	timespec_add(&tspecs.expected, &period_spec);
	timespec_add(&tspecs.last, &period_spec);

	mono_time(&t_mono);
	t_ns = mono_time_ns(&t_mono);
	resp_time = ts_mono_diff(&tspecs.rel, &t_mono);
	__jitter_rec(t_ctx, THREAD_DISPLAY, PROF_RESPONSE, disp_frame_num,
			resp_time, t_ns);

	/* okay now increment release time */
	timespec_add(&tspecs.rel, &period_spec);

}

static inline void display_done(struct thread_ctx* t_ctx)
{
#ifdef USE_LITMUS
	/* LITMUS just sleeps for the next period */
	litmus_job_completion();

#else /* USE_LITMUS */
#ifdef __APPLE__
#error No clock_nanosleep(2) on Apple. Need to implement timespec subtraction.
#else /* __APPLE__ (within else of USE_LITMUS)  */

	int err;

	/* non-LITMUS must sleep a bit */
	err = clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME,
			&tspecs.rel, NULL);
	if (err)
	{
		if (err == EINTR && should_quit())
			fprintf(stderr, "quitting (nanosleep interrupted)\n");
		else
			die("dying after bad clock_nanosleep!");
	}
#endif /* __APPLE__ */
#endif /* USE_LITMUS */
}

/*
 * Called after the call to SDL returns to record jitter.
 *
 * Pre-Condition: expected_ts has the time we should have shown the frame, and
 * last_ts has the time that we showd the last frame, plus a period.
 *
 * Post-Condition: last_ts is equal to the current time used to record jitter.
 *
 * The timespecs are incremented by a period in another method.
 */
static inline void record_jitter(struct thread_ctx *t_ctx)
{
	mono_time_t cur_mono;

	mono_time(&cur_mono);

	if (tspec_uninit(&tspecs.expected))
	{
		/* the first frame sets the start time */
		mono_time_timespec(&tspecs.expected, &cur_mono);
		mono_time_timespec(&tspecs.last, &cur_mono);
	}

	/* save the jitters */
	prof_jitter_rec(t_ctx, THREAD_DISPLAY, PROF_JITTER_ABS, disp_frame_num,
			&tspecs.expected, &cur_mono);
	prof_jitter_rec(t_ctx, THREAD_DISPLAY, PROF_JITTER_RELATIVE,
			disp_frame_num, &tspecs.last, &cur_mono);

	/* establish the post condition */
	mono_time_timespec(&tspecs.last, &cur_mono);
}

static void read_from_queue(struct thread_ctx *thread_ctx)
{
	struct video_state *vs = thread_ctx->vs;
	struct sems *sems = thread_ctx->sems;
	int sem_val, err;

	/*
	 * There is a race condition here, whose workaround is hackey and should
	 * be fixed.
	 *
	 * If the decoder decodes the last frame, posts to vidq_empty, and then
	 * stalls before setting decoder_done, we could wait on the semaphore in
	 * the else case below when we don't need to. As a fix, the decoder will
	 * post to the semaphore after setting decoder_done to wake us up.
	 *
	 * In the case that the race condition does not occur (normal case), the
	 * semaphore could be one higher than the true queue count.  However,
	 * this is detected by a failed queue_read() when decoder_done is set.
	 */

	if (thread_ctx->decoder_done)
	{
		/* The decoder is done, don't block on the semaphore.
		 * NOTE that these two calls can't be ANDed in the if, or we
		 * decrement the semaphore twice!
		 */
		if (-1 == sem_trywait(sems->vidq_empty)) {
			/* Decoder done and the video queue is empty == there are
			 * no more frames to display, so quit. */
			set_should_quit();
			return;
		}
	} else {
		err = sem_getvalue(sems->vidq_empty, &sem_val);
		if (err)
			die("Bad sem_getvalue");
		prof_buffer_rec(thread_ctx, THREAD_DISPLAY,
				disp_frame_num + 1, sem_val);

		/* Decoder not done. Use normal blocking semaphore calls
		 * also the frame number has not yet been incremented here */
		sem_wait_wrap(sems->vidq_empty, thread_ctx, THREAD_DISPLAY,
				disp_frame_num + 1);
	}

	/* copy the next vid_q_item into our memory space */
	if (!queue_read(vs->vid_q, &src_vq_item))
	{
		if (!thread_ctx->decoder_done)
		{
			/* should not happen if semaphores are correct! */
			fprintf(stderr, "ERROR: Display queue read. DO NOT IGNORE!");
		} else {
			set_should_quit();
			return;
		}
	}

	/* Successfully read from queue. Corresponding wait is in the video
	 * decode loop */
	sem_post_wrap(sems->vidq_full);
}

/* Actually put the frame on the screen */
static inline void draw_frame(struct thread_ctx *t_ctx)
{
	const int width = t_ctx->vs->codec_ctx->width;
	const int height = t_ctx->vs->codec_ctx->height;

	if (t_ctx->opts->use_sdl)
		SDL_LockSurface(screen);

	PROFILE_VOID_FUNCTION_CALL(t_ctx, THREAD_DISPLAY, PROF_PICT_COPY,
			av_picture_copy, &screen_pict, &src_vq_item.picture,
			SCREEN_PIX_FMT, width, height);

	if (t_ctx->opts->use_sdl)
		SDL_UnlockSurface(screen);

	SDL_UpdateRect(screen, 0, 0, 0, 0);
	++disp_frame_num;
}

static void ready_first_frame(struct thread_ctx *t_ctx)
{
#ifdef USE_LITMUS
	int err, sem_val;

	err = sem_getvalue(t_ctx->sems->vidq_empty, &sem_val);
	if (err)
		die("Could not check vidq_empty value");

	if (sem_val < 1)
		die("At least one frame must be buffered under LITMUS");
#endif

	read_from_queue(t_ctx);
}

static inline void release_jitter(struct thread_ctx *t_ctx)
{
	mono_time_t now;
	mono_time(&now);

	if (tspec_uninit(&tspecs.rel))
		mono_time_timespec(&tspecs.rel, &now);

	prof_jitter_rec(t_ctx, THREAD_DISPLAY, PROF_JITTER_RELEASE,
			disp_frame_num + 1, &tspecs.rel, &now);
}

void video_display_loop(struct thread_ctx *thread_ctx)
{
#ifdef USE_PROFILE
	unsigned long long start, end;
#endif

	wait_preload_queue();

	ready_first_frame(thread_ctx);

	/* Only LITMUS^RT */
	make_litmus_display(thread_ctx);
	wait_for_release();

	while (!should_quit()) {

#ifdef USE_PROFILE
		start = cpu_time();
#endif

		release_jitter(thread_ctx);
		draw_frame(thread_ctx);
		record_jitter(thread_ctx);
		read_from_queue(thread_ctx);
		update_timespecs(thread_ctx);

#ifdef USE_PROFILE
		end = cpu_time();
		prof_start_end(thread_ctx, THREAD_DISPLAY, PROF_FULL_DISPLAY,
				start, end);
#endif

		display_done(thread_ctx);
	}
	become_regular_task();
}

void tear_down_display(struct options* opts)
{
	if (!opts->use_sdl)
		avpicture_free(&screen_pict);
	avpicture_free(&src_vq_item.picture);
}
