#ifndef _TIMING_H
#define _TIMING_H

#include <unistd.h>

#define USE_POSIX_TIMERS (defined(_POSIX_TIMERS) && _POSIX_TIMERS > 0 && defined(_POSIX_MONOTONIC_CLOCK))
#define USE_APPLE_TIMERS (__APPLE__)

#if USE_POSIX_TIMERS
#include <time.h>
typedef struct timespec mono_time_t;
#define MSEC_PER_SEC 1000
#define USEC_PER_SEC 1e6
#define NSEC_PER_SEC 1e9
#elif USE_APPLE_TIMERS
#include <mach/mach_init.h>
#include <mach/mach_time.h>
#include <mach/thread_info.h>
#include <mach/thread_act.h>
typedef uint64_t mono_time_t;
#else
#error no good monotonic type found
#endif

#include "common.h"

unsigned long long cpu_time(void);
void initialize_mono_time_t(mono_time_t*);

static inline void mono_time(mono_time_t*);
static inline void mono_time_timespec(struct timespec*, const mono_time_t*);
static inline unsigned long long mono_time_ns(const mono_time_t*);
static inline int mono_time_t_is_zero(mono_time_t*);

static inline unsigned long long mono_time_diff_nsec(mono_time_t*,
		mono_time_t*);
static inline double mono_time_diff_sec(mono_time_t*, mono_time_t*);
static inline void timespec_add(struct timespec*, const struct timespec*);

#if USE_POSIX_TIMERS
static inline void mono_time(mono_time_t *ts)
{
	int err;
	err = clock_gettime(CLOCK_MONOTONIC, ts);
	if (err != 0)
	{
		perror("clock_gettime type MONOTONIC RAW");
		exit(EXIT_FAILURE);
	}
}

static inline void mono_time_timespec(struct timespec *ts,
		const mono_time_t *mt)
{
	/* In linux, mono_time_t is already a timespec */
	ts->tv_sec	= mt->tv_sec;
	ts->tv_nsec	= mt->tv_nsec;
}

static inline unsigned long long mono_time_ns(const mono_time_t *t)
{
	return (((unsigned long long)t->tv_sec) * NSEC_PER_SEC +
			((unsigned long long)t->tv_nsec));
}

static inline int mono_time_t_is_zero(mono_time_t* ts)
{
	return (ts->tv_sec == 0 && ts->tv_nsec == 0);
}

#elif USE_APPLE_TIMERS

/*
 * See http://developer.apple.com/library/mac/#qa/qa2004/qa1398.html
 */
void mono_time(mono_time_t *t)
{
	*t = mach_absolute_time();
}

static inline void mono_time_timespec(struct timespec *ts,
		const mono_time_t * mt)
{
	ts->tv_sec  = *mt / NSEC_PER_SEC;
	ts->tv_nsec = *mt - ((unsigned long long)ts->tv_sec) * NSEC_PER_SEC;
}

static inline int mono_time_t_is_zero(mono_time_t *t)
{
	return (*t == 0);
}

static unsigned long long mono_time_ns(mono_time_t *t)
{
	static mach_timebase_info_data_t tb_info = { 0, 0 };
	static kern_return_t err;

	/* this is only done once (the info structure is static) */
	if (tb_info.denom == 0)
	{
		err = mach_timebase_info(&tb_info);
		if (err != KERN_SUCCESS)
		{
			fprintf(stderr, "call to timebase_info failed\n");
			exit(EXIT_FAILURE);
		}
	}
	return (*t * tb_info.numer / tb_info.denom);
}
#endif /* USE_{POSIX or APPLE}_TIMERS */

static inline unsigned long long mono_time_diff_nsec(mono_time_t *start_in,
		mono_time_t *end_in)
{
	const unsigned long long start	= mono_time_ns(start_in);
	const unsigned long long end	= mono_time_ns(end_in);

	/* overflow check */
	if (start > end)
		die("Called mono time diff with start > end");
	return (end - start);
}

static inline double mono_time_diff_sec(mono_time_t *start, mono_time_t *end)
{
	const unsigned long long diff_nsec = mono_time_diff_nsec(start, end);
	return (((double)diff_nsec) / NSEC_PER_SEC);
}

static inline void timespec_add(struct timespec *res,
		const struct timespec *other)
{
	const long nsec_sum	= res->tv_nsec + other->tv_nsec;
	const long sec_carry	= nsec_sum / NSEC_PER_SEC;
	res->tv_sec		= res->tv_sec + other->tv_sec + sec_carry;
	res->tv_nsec		= nsec_sum - (sec_carry * NSEC_PER_SEC);

}

/*
 * Convert a timespect to nanoseconds in unsigned long long
 */
static inline unsigned long long ts_to_ns(const struct timespec *ts)
{
	return ( ((unsigned long long)ts->tv_sec) * NSEC_PER_SEC +
			ts->tv_nsec );
}

static inline int tspec_uninit(const struct timespec *ts)
{
	return (ts->tv_sec == 0 && ts->tv_nsec == 0);
}

/*
 * Calculate the difference between the expected time and what it is now.
 */
static inline long long ts_mono_diff(const struct timespec *expected_ts,
		const mono_time_t *now_mono)
{
	/* use longs so we can get negative numbers */
	const long long now_ns = mono_time_ns(now_mono);
	const long long expected_ns = ts_to_ns(expected_ts);
	return (now_ns - expected_ns);
}
#endif /* _TIMING_H */
