#include <pthread.h>
#include <signal.h>
#include <sched.h>

#include <sys/resource.h>

/*
 * If using SDL, the file containing the main() function must include SDL.h
 */
#include <SDL.h>

#include "inc/video_common.h"
#include "inc/options.h"
#include "inc/timing.h"
#include "inc/common.h"
#include "inc/display_common.h"
#include "inc/thread_ctx.h"
#include "inc/sem_common.h"
#include "inc/profile.h"
#include "inc/video_queue.h"

static int quit_flag = 0;
pthread_t decode_thread;

static void usage(char *prog_name)
{
	fprintf(stderr, "Usage: %s [OPTIONS] <file>\n", prog_name);
	fprintf(stderr, "OPTIONS:\n"
			"    -s  Don't use SDL (fake display)\n"
#ifndef USE_LITMUS
		        "    -f  Schedule with SCHED_FIFO\n"
#endif
			"    -b  Budget for decode task (milliseconds)\n"
			"    -c  CPU for the HRT server to run on (LITMUS)\n"
			"    -r  Queue preloading time\n"
			"    -t  Time to stop playing video after\n"
			"    -k  Seek to keyframe nearest this time\n"
			"    -p  Set the period (override what file says)\n"
			"    -h  Set the phase\n"
			"All times are in milliseconds.\n");
}


static void output_header(struct video_state* vs)
{
	char buf[256];

	avcodec_string(buf, sizeof(buf),
			vs->fmt_ctx->streams[vs->video_stream]->codec, 0);
	printf("%s\n", buf);

	/* keep track which version was used to obtain experimental data */
	printf("git=%s\n", GIT_HEAD);
}

int should_quit(void)
{
	return quit_flag;
}

void set_should_quit(void)
{
	quit_flag = 1;
}

static void int_handler(int signum)
{
	fprintf(stderr, "Interrupted. Cleaning up...\n");
	set_should_quit();
}


#ifndef USE_LITMUS
/* FIFO scheduling priority under linux */
#define DISPLAY_FIFO_PRIO sched_get_priority_max(SCHED_FIFO)
#define DECODE_FIFO_PRIO  (DISPLAY_FIFO_PRIO - 1)
void fifo_set_display_prio(struct options *opts){
	/* display is this main process */
	const struct sched_param param = {
		.sched_priority = DISPLAY_FIFO_PRIO,
	};
	int err;

	if (opts->use_fifo) {
		err = sched_setscheduler(0, SCHED_FIFO, &param);
		if (err)
			die("Could not set display to SCHED_FIFO");
	}
}

void fifo_set_decode_prio(struct options *opts){
	const struct sched_param param = {
		.sched_priority = DECODE_FIFO_PRIO,
	};

	int err;

	if (opts->use_fifo){
		err = pthread_setschedparam(decode_thread, SCHED_FIFO, &param);
		if (err)
			die("Could not set decoder to SCHED_FIFO");
	}
}
#else
#define fifo_set_display_prio(opts)	do { } while (0)
#define fifo_set_decode_prio(opts)	do { } while (0)
#endif

int main (int argc, char **argv)
{
	struct sigaction sig_int = { .sa_handler = int_handler, .sa_flags = 0 };
	struct thread_ctx thread_ctx;
	struct video_state *vs;
	struct options opts;
	struct sems sems;
	sigset_t sigset;
	void *join_res = NULL;
	int err;

	sigemptyset(&sigset);
	sig_int.sa_mask = sigset;

	if (sigaction(SIGINT, &sig_int, 0) || sigaction(SIGTERM, &sig_int, 0))
		die("Could not install interrupt signal handler\n");

	url_set_interrupt_cb(should_quit);

	vs = av_malloc(sizeof(*vs));
	if (!vs)
		die("Could not allocate memory\n");

	err = parse_args(vs, &opts, argc, argv);
	if (err) {
		usage(argv[0]);
		die("Invalid arguments\n");
	}

	if (opts.use_sdl) {
		if (SDL_Init(SDL_INIT_VIDEO))
			die("could not initialize sdl\n");
	}

	init_thread_ctx(&thread_ctx, vs, &opts, &sems);
	initialize_av(&thread_ctx);
	initialize_display(&thread_ctx);
	initialize_vid_q(&thread_ctx);
	initialize_sems(&thread_ctx);

	output_header(vs);

	err = pthread_create(&decode_thread, NULL, decode_thread_loop,
			(void*) &thread_ctx);
	if (err)
		die("Could not start decode thread\n");

	fifo_set_display_prio(&opts);
	fifo_set_decode_prio(&opts);

	video_display_loop(&thread_ctx);

	err = pthread_cancel(decode_thread);
	if (err && err != ESRCH)
		fprintf(stderr, "pthread_cancel returned an error\n");

	prof_completion_event(&thread_ctx, THREAD_DISPLAY, cpu_time(),
			vs->frames_decoded);

	err = pthread_join(decode_thread, &join_res);
	if (err && err != ESRCH)
		fprintf(stderr, "Could not join decoder thread\n");

	if (join_res == PTHREAD_CANCELED)
		fprintf(stderr, "Decode thread was canceled\n");
	else if ( *((int*)join_res) == 1 )
		fprintf(stderr, "Decode thread exited normally\n");
	else
		fprintf(stderr, "Decoded thread exited abnormally\n");

	destroy_sems(thread_ctx.sems);
	tear_down_vid_q(vs);
	tear_down_display(&opts);
	tear_down_av(vs);
	thread_save_trace_buffers(&thread_ctx);
	destroy_thread_ctx(&thread_ctx);

	printf("frames decoded: %llu\n", vs->frames_decoded);

	av_freep(&vs);
	SDL_Quit();

	exit(EXIT_SUCCESS);
}
