#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <math.h>

#include "inc/video_common.h"
#include "inc/common.h"
#include "inc/options.h"

#if defined(linux) || defined(__linux) || defined(__linux__)
#if _XOPEN_SOURCE >= 600 || _ISOC99_SOURCE || _POSIX_C_SOURCE >= 200112L
/* Linux is okay */
#else
#error no suitable floating point features
#endif /* Linux */
#elif defined(__APPLE__)
/* fpclassify is in math.h, no checks needed*/
#else
#error Unrecognized OS
#endif /* OS Check */

static double xstrtod(const char* str_in)
{
	char *endptr;
	double val;
	int class;

	errno = 0;
	val = strtod(str_in, &endptr);

	if ((val == HUGE_VALF || val == HUGE_VALL) && errno != 0)
	{
		perror("strtod");
		die("bad strtod");
	}

	if (str_in == endptr)
		die("No value converted in strtod");

	class= fpclassify(val);
	if (class == FP_NAN || class == FP_INFINITE)
		die("Can't use NAN or infinite numbers");

	return val;
}

static long xstrtol(const char* str_in)
{
	char *endptr;
	long int val;

	errno = 0;
	val = strtol(str_in, &endptr, 10);

	if ((errno == ERANGE && (val == LONG_MAX || val == LONG_MIN))
			|| (errno != 0 && val == 0)) {
		die("strtol");
		die("bad strtol");
	}

	if (endptr == str_in)
		die("no value found in strtol");

	return val;
}

#ifdef USE_LITMUS
#define OPTSTR "sb:c:r:t:k:p:h:"
#else
#define OPTSTR "sfb:c:r:t:k:p:h:"
#endif

int parse_args(struct video_state *vs, struct options *opts,
		int argc, char **argv)
{
	int opt;

	/* set default options */
	opts->use_sdl = 1;
	opts->budget = 0.0;
	opts->preload_time = 0.0;
	opts->cpu = -1;
	opts->stop_after = 0.0;
	opts->seek_to = 0.0;
	opts->set_period = 0.0;
	opts->phase = -1;
#ifndef USE_LITMUS
	opts->use_fifo = 0;
#endif

	while ((opt = getopt(argc, argv, OPTSTR)) != -1) {
		switch (opt) {
		case 's':
			opts->use_sdl = 0;
			break;
#ifndef USE_LITMUS
		case 'f':
			opts->use_fifo = 1;
			break;
#endif
		case 'b':
			opts->budget = xstrtod(optarg);
			break;
		case 'c':
			opts->cpu = xstrtol(optarg);
			break;
		case 'r':
			opts->preload_time = xstrtod(optarg);
			break;
		case 't':
			opts->stop_after = xstrtod(optarg);
			break;
		case 'k':
			opts->seek_to = xstrtod(optarg);
			break;
		case 'p':
			opts->set_period = xstrtod(optarg);
			break;
		case 'h':
			opts->phase = xstrtod(optarg);
			break;
		case ':':
			return -1;
		case '?':
		default:
			return -1;
		}
	}

	if (optind < argc)
		vs->fname = argv[optind];
	else {
		fprintf(stderr, "Need a file\n");
		return -1;
	}

	if (opts->budget < 0)
		die("Preload time cannot be < 0");

	if (opts->stop_after < 0)
		die("The stop-after time cannot be negative");

	if (opts->seek_to < 0)
		die("The seek-to time cannot be negative");

	if (opts->set_period < 0.0)
		die("Can't have a negative period");

#ifdef USE_LITMUS
	if (opts->budget == 0.0)
		die("LITMUS needs a budget");
	if (opts->cpu == -1)
		die("LITMUS needs a CPU for the HRT");
	if (opts->preload_time == 0.0)
		die("LITMUS needs some preloading time");
	if (opts->phase == -1)
		die("LITMUS needs a phase");
#endif

	return 0;
}
