#include "inc/timing.h"
#include "inc/thread_ctx.h"
#include "inc/profile.h"
#include "inc/common.h"

/*
 * Increment record selector and return the location of the next slot.
 */
int next_record_idx(struct thread_ctx *t_ctx,
		const enum thread_selector thread)
{
	const int old_val = fetch_and_inc(&t_ctx->rec_buf_idx[thread]);
	return old_val;
}

static void __save_buf_overflow(struct thread_ctx *t_ctx,
		const enum thread_selector thread,
		const int buffer_idx)
{
	struct prof_event_record *rec = &t_ctx->rec_buf[thread][buffer_idx];
	rec->hdr.type = PROF_OVERFLOW;
}

static void __prof_rec_buf_overflow(struct thread_ctx *t_ctx,
		const enum thread_selector this_thread,
		const int this_idx)
{
	const enum thread_selector other_thread =
		(this_thread == THREAD_DISPLAY) ?  THREAD_DECODE : THREAD_DISPLAY;

	int other_idx = next_record_idx(t_ctx, other_thread);
	
	if (other_idx >= THREAD_REC_BUF_COUNT)
	{
		/* use the 2nd to last position, in case the other thread is
		 * going to write in the last position */
		other_idx = THREAD_REC_BUF_COUNT - 2;
		if (other_idx < 0)
			other_idx = 0;
	}

	__save_buf_overflow(t_ctx, this_thread, this_idx);
	__save_buf_overflow(t_ctx, other_thread, other_idx);
	thread_save_trace_buffers(t_ctx);
	die("Buffer overflowed!");
}

static void prof_rec_buf_overflow(struct thread_ctx *t_ctx,
		const enum thread_selector thread_idx_known,
		const int known_idx)
{
	static int guard = 0;
	const int skip = fetch_and_inc(&guard);

	/* only one thread should proceed */
	if (!skip)
		__prof_rec_buf_overflow(t_ctx, thread_idx_known, known_idx);
}

/*
 * Get point to next buffer record. Emergency exit if needed.
 */
static struct prof_event_record* next_rec_buf(struct thread_ctx *t_ctx,
		const enum thread_selector thread)
{
	const int next_idx = next_record_idx(t_ctx, thread);
	struct prof_event_record *rec_ret = &t_ctx->rec_buf[thread][next_idx];

	if (next_idx == THREAD_REC_BUF_COUNT - 1)
	{
		/* last record, emergency exit */
		prof_rec_buf_overflow(t_ctx, thread, next_idx);
		return NULL;
	} else
		return rec_ret;
}

void prof_completion_event(
		struct thread_ctx *t_ctx,
		const enum thread_selector thread,
		const unsigned long long total_time,
		const unsigned long long frames_decoded)
{
	struct prof_event_record *rec = next_rec_buf(t_ctx, thread);
	const pid_t pid = getpid();

	rec->hdr.type				= PROF_COMPLETION;
	rec->data.completion.total_time		= total_time;
	rec->data.completion.frames_decoded	= frames_decoded;
	rec->data.completion.pid_group		= pid;
}

/*
 * Record the jitter with the type and values passed in.
 */
void __jitter_rec(
		struct thread_ctx		*t_ctx,
		const enum thread_selector	thread,
		const prof_record_type_t	type,
		const unsigned long long	frame,
		const long long			jitter,
		const unsigned long long	time)
{
	struct prof_event_record *rec = next_rec_buf(t_ctx, thread);
	rec->hdr.type		= type;
	rec->data.jitter.frame	= frame;
	rec->data.jitter.jitter	= jitter;
	rec->data.jitter.time	= time;
}

/*
 * Record the jitter with the type passed in. The expected_ts is the time we
 * expected to see this record, and the cur_ns is the time that it is
 * presently.
 */
void prof_jitter_rec(
		struct thread_ctx		*t_ctx,
		const enum thread_selector	thread,
		const prof_record_type_t	type,
		const unsigned long long	frame,
		const struct timespec		*expected_ts,
		const mono_time_t		*now_mono)
{
	const long long jitter	= ts_mono_diff(expected_ts, now_mono);
	const long long time	= ts_to_ns(now_mono);
	__jitter_rec(t_ctx, thread, type, frame, jitter, time);
}

void prof_buffer_rec(
		struct thread_ctx		*t_ctx,
		const enum thread_selector	thread,
		const unsigned long long	frame,
		const int			sem_val)
{
	struct prof_event_record *rec = next_rec_buf(t_ctx, thread);
	mono_time_t time;

	mono_time(&time);
	rec->hdr.type		= PROF_BUFFER;
	rec->data.buffer.frame	= frame;
	rec->data.buffer.time	= mono_time_ns(&time);
	rec->data.buffer.queued	= sem_val;
}

#ifdef USE_PROFILE
void prof_start_end(struct thread_ctx *t_ctx,
		const enum thread_selector thread,
		prof_record_type_t rec_type,
		unsigned long long start, unsigned long long end)
{
	struct prof_event_record *rec	= next_rec_buf(t_ctx, thread);
	rec->hdr.type			= rec_type;
	rec->data.start_end.start	= start;
	rec->data.start_end.end		= end;
}

void prof_sem_wait(struct thread_ctx *t_ctx, const enum thread_selector thread,
		unsigned long long frame)
{
	struct prof_event_record *rec	= next_rec_buf(t_ctx, thread);
	rec->hdr.type			= PROF_SEM_WAIT;
	rec->data.sem_wait.frame	= frame;
}
#endif
