#include <limits.h>
#include <libavutil/avutil.h>

#include "inc/common.h"
#include "inc/queue.h"

enum slot_status {
	SLOT_FREE,
	SLOT_BUSY,
	SLOT_READY
};

struct queue {
	unsigned int	slot_count;
	unsigned int	slot_size;

	int 		free_count;
	unsigned int 	write_idx;
	unsigned int 	read_idx;

	char*		slots;
	void*		buffer_mem;
	unsigned int	failed_writes;

	void		(*cpy_fcn)(void*, void*);
};

int queue_init(	struct queue**	q_in,
		unsigned int 	slot_count,
		unsigned int 	slot_size,
		char*		slots,
		void* 		buffer_mem,
		void		(*cpy_fcn)(void*, void*))
{
	int i = 0;
	struct queue* q;

	if (!slot_count || UINT_MAX % slot_count != slot_count - 1) {
		/* The slot count must divide UNIT_MAX + 1 so that when it
		 * wraps around the index correctly points to 0.
		 * Also, remember UINT_MAX = (2^32 - 1), not 2^32.
		 */
		*q_in = NULL;
		return 0;
	} else {
		q = av_malloc(sizeof(*q));
		if (!q)
			die("Could not allocate queue memory\n");
		*q_in = q;
		q->slot_count		= slot_count;
		q->slot_size		= slot_size;
		q->slots		= slots;
		q->buffer_mem		= buffer_mem;
		q->free_count		= slot_count;
		q->write_idx		= 0;
		q->read_idx		= 0;
		q->failed_writes	= 0;
		q->cpy_fcn		= cpy_fcn;
		for (i = 0; i < slot_count; i++)
			q->slots[i] = (enum slot_status) SLOT_FREE;
		return 1;
	}
}

int queue_start_write(struct queue* q, void **ptr)
{
	int free = fetch_and_dec(&q->free_count);
	unsigned int idx;
	if (free <= 0) {
		fetch_and_inc(&q->free_count);
		*ptr = NULL;
		fetch_and_inc((int*) &q->failed_writes);
		return 0;
	} else {
		idx  = fetch_and_inc((int*) &q->write_idx) % q->slot_count;
		q->slots[idx] = (enum slot_status) SLOT_BUSY;
		*ptr = ((char*) q->buffer_mem) + idx * q->slot_size;
		return 1;
	}
}

void queue_finish_write(struct queue* q, void *ptr)
{
	unsigned int idx = ((char*) ptr - (char*) q->buffer_mem) / q->slot_size;
	q->slots[idx]	 = (enum slot_status) SLOT_READY;
}

/* exclusive reader access is assumed */
int queue_read(struct queue* q, void* dest)
{
	unsigned int idx;
	if (q->free_count == q->slot_count)
		/* nothing available */
		return 0;
	idx = q->read_idx % q->slot_count;
	if (q->slots[idx] == (enum slot_status) SLOT_READY) {
		q->cpy_fcn(dest, ((char*) q->buffer_mem) + idx * q->slot_size);
		q->slots[idx] = (enum slot_status) SLOT_FREE;
		q->read_idx++;
		fetch_and_inc(&q->free_count);
		return 1;
	} else
		return 0;
}

void queue_destroy(struct queue** q_in)
{
	struct queue* q;

	if (q_in == NULL)
		return;

	q = *q_in;
	av_free(q->buffer_mem);
	av_free(q->slots);
	av_freep(q_in);
}
