#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>

#include "inc/options.h"
#include "inc/thread_ctx.h"
#include "inc/video_queue.h"
#include "inc/sem_common.h"

/*
 * Only evaluate the arguments a and b once
 */
#define MAX(a,b) \
       ({ typeof (a) _a = (a); \
           typeof (b) _b = (b); \
         _a > _b ? _a : _b; })

#define SEM_VIDQ_FULL	"/video_queue_full"
#define SEM_VIDQ_EMPTY	"/video_queue_empty"

/*
 * This is enough to represent a 32-bit int as a string...
 */
#define PID_STR_LEN	10

/* Note that sizeof on the SEM_VIDQ_{FULL,EMPTY} includes the null byte */
#define SEM_NAME_MAX_LEN (MAX(sizeof(SEM_VIDQ_FULL),sizeof(SEM_VIDQ_EMPTY)) \
		+ PID_STR_LEN)


void initialize_sems(struct thread_ctx* t_ctx)
{
	const pid_t my_pid = getpid(); /* getpid always works (imagine that) */
	struct sems* sems = t_ctx->sems;
	int n_chars, max_queue_size;

	max_queue_size = frames_to_preload(t_ctx);
	if (max_queue_size < 1)
	{
		printf("Setting queue size to 1 when preload was 0\n");
		max_queue_size = 1;
	}

	if (max_queue_size > t_ctx->queue_size)
		die("max queue size > queue size");

	sems->vidq_name_full = av_malloc(SEM_NAME_MAX_LEN);
	sems->vidq_name_empty = av_malloc(SEM_NAME_MAX_LEN);

	if (!sems->vidq_name_full || !sems->vidq_name_empty)
		die("Couldn't allocate semaphore name strings\n");

	n_chars = snprintf(sems->vidq_name_full, SEM_NAME_MAX_LEN,
			"%s%d", SEM_VIDQ_FULL, my_pid);
	if (n_chars < 1)
		die("Couldn't construct semaphore name\n");

	n_chars = snprintf(sems->vidq_name_empty, SEM_NAME_MAX_LEN,
			"%s%d", SEM_VIDQ_EMPTY, my_pid);
	if (n_chars < 1)
		die("Couldn't construct semaphore name\n");

	sems->vidq_full = sem_new(sems->vidq_name_full, max_queue_size);
	sems->vidq_empty = sem_new(sems->vidq_name_empty, 0);
}

sem_t *sem_new(const char* name, int value)
{
	sem_t *sem;
	int err;

	errno = 0;
	sem = sem_open(name, O_CREAT|O_EXCL, S_IRUSR|S_IWUSR, value);

	if (SEM_FAILED == sem)
	{
		fprintf(stderr,  "Could not open a semaphore...\n");

		if (EEXIST == errno)
		{
			fprintf(stderr, "It existed. Trying to delete it...\n");

			err = sem_unlink(name);
			if (err)
				die("Could not unlink semaphore. I give up.\n");

			fprintf(stderr, "Unlinked it. Trying again...\n");

			sem = sem_open(name, O_CREAT|O_EXCL,
					S_IRUSR|S_IWUSR, value);
			if (SEM_FAILED == sem)
				die("Could not re-create. Something is very wrong\n");

			fprintf(stderr, "Everything seems okay now\n");
		} else {
			fprintf(stderr, "Some other error. I give up.\n");
			die("Semaphores are messed up!\n");
		}
	}
	return sem;
}

static void sem_close_wrap(sem_t *sem)
{
	if (sem_close(sem))
		fprintf(stderr, "Could not close semaphore\n");
}

static void sem_unlink_wrap(const char *name)
{
	if (sem_unlink(name))
		fprintf(stderr, "Could not unlink semaphore\n");
}

void destroy_sems(struct sems* sems)
{
	sem_close_wrap(sems->vidq_full);
	sem_close_wrap(sems->vidq_empty);
	sem_unlink_wrap(sems->vidq_name_full);
	sem_unlink_wrap(sems->vidq_name_empty);
	av_free(sems->vidq_name_full);
	av_free(sems->vidq_name_empty);
}
