#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <sys/utsname.h>
#include <pthread.h>

#include "inc/common.h"
#include "inc/profile.h"
#include "inc/video_common.h"
#include "inc/thread_ctx.h"

#define FNAME_START	"myplayer"


/* to signal the display it's okay to proceed */
static struct preload_queue_sync {
	int		done;
	pthread_mutex_t	mx;
	pthread_cond_t	cond;
} preloaded_queue_sync;

/*
 * Get thread output file name.
 *
 * The filname must:
 *	- Start with myplayer (and an underscore)
 *	- Contain the period
 *	- Contain the video-file name
 *
 * Arguments:
 *	@thread_type is used to indicate decode or display
 *	@t_ctx is the thread context structure.
 *	@name is a buffer where we should write the name into.
 *	@maxlen is the size of the @name buffer.
 */
static char* get_std_name(const char* thread_type,
		const struct thread_ctx* t_ctx, char *name, int maxlen)
{
	struct utsname uts;
	char *p, *buf, *new_fname;
	double period_msec = t_ctx->period / ((double)1e6);

	buf = new_fname = strdup(t_ctx->vs->fname);
	for (p = new_fname; *p; p++)
		if (*p == ' ' || *p == '_')
			*p = '-';
	/* remove prefix path */
	for (; p > new_fname; p--)
		if (*p == '/')
			new_fname = p + 1;

	if (0 != uname(&uts))
		snprintf(uts.nodename, sizeof(uts.nodename), "unknown");

	snprintf(name, maxlen,
			FNAME_START "_" "host=%s" "_" "vid-file=%s" "_"
			"period=%.6f" "_" "thread=%s.bin",
			uts.nodename, new_fname, period_msec,
			thread_type);

	free(buf);

	return name;
}

static int xopen(const char* fname)
{
	int fd = open(fname, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (0 > fd)
		die("Could not open file for decode data\n");
	return fd;
}

static void setup_preload_queue_cond(void)
{
	int err;

	preloaded_queue_sync.done = 0;

	err = pthread_cond_init(&preloaded_queue_sync.cond, NULL);
	if (err)
		die("Could not create preload condition variable");

	err = pthread_mutex_init(&preloaded_queue_sync.mx, NULL);
	if (err)
		die("Could not create preload mutex");
}

static void init_record_buffers(struct thread_ctx *t_ctx)
{
	struct prof_event_record *new_rec_buf;
	int thread;

	for (thread = ((enum thread_selector) 0);
			thread < THREAD_COUNT; thread++)
	{
		new_rec_buf = av_malloc(sizeof(*new_rec_buf) *
				THREAD_REC_BUF_COUNT);
		if (!new_rec_buf)
			die("Could not allocate record buffer");

		t_ctx->rec_buf[thread] = new_rec_buf;
		t_ctx->rec_buf_idx[thread] = 0;
	}
}

void init_thread_ctx(struct thread_ctx *t_ctx, struct video_state *vs,
		struct options *opts, struct sems *sems)
{
	t_ctx->vs = vs;
	t_ctx->opts = opts;
	t_ctx->sems = sems;
	t_ctx->decoder_done = 0;
	t_ctx->queue_size = 0;
	init_record_buffers(t_ctx);
	setup_preload_queue_cond();
}

void wait_preload_queue()
{
	int err;
	pthread_mutex_t *const mx = &preloaded_queue_sync.mx;
	pthread_cond_t *const cond = &preloaded_queue_sync.cond;

	pthread_mutex_lock(mx);
	while (!preloaded_queue_sync.done)
	{
		printf("Queue not preloaded yet...\n");
		err = pthread_cond_wait(cond, mx);
		if (err)
			die("Error on cond_wait");
	}
	pthread_mutex_unlock(mx);
	printf("Queue preloaded. Continuing...\n");
}

/* signal the display that it is safe to proceed */
void signal_preload_queue()
{
	pthread_mutex_t *const mx = &preloaded_queue_sync.mx;
	pthread_mutex_lock(mx);
	preloaded_queue_sync.done = 1;
	pthread_cond_signal(&preloaded_queue_sync.cond);
	pthread_mutex_unlock(mx);
}

static void destroy_preload_queue_cond(struct thread_ctx *t_ctx)
{
	int err;
	err = pthread_mutex_destroy(&preloaded_queue_sync.mx);
	err = err || pthread_cond_destroy(&preloaded_queue_sync.cond);
	if (err)
		fprintf(stderr, "Couldn't destroy preload mutex or cond?\n");
}

void destroy_thread_ctx(struct thread_ctx *t_ctx)
{
	destroy_preload_queue_cond(t_ctx);
}

static void __thread_save_trace_buffers(struct thread_ctx *t_ctx,
		const enum thread_selector thread,
		const char* thread_name)
{
	const size_t to_write = sizeof(struct prof_event_record);
	struct prof_event_record *rec = t_ctx->rec_buf[thread];
	char buf[PATH_MAX];
	char *fname = NULL;
	int last_idx, fd, i;
	ssize_t written;

	fname = get_std_name(thread_name, t_ctx, buf, sizeof(buf));
	fd = xopen(fname);

	last_idx = next_record_idx(t_ctx, thread);
	if (last_idx >= THREAD_REC_BUF_COUNT)
		last_idx = THREAD_REC_BUF_COUNT;

	for (i = 0; i < last_idx; ++i, ++rec)
	{
		written = write(fd, rec, to_write);
		if (written != to_write)
			die("Incomplete trace write!");
	}
	close(fd);
}

void thread_save_trace_buffers(struct thread_ctx *t_ctx)
{
	__thread_save_trace_buffers(t_ctx, THREAD_DECODE, "decode");
	__thread_save_trace_buffers(t_ctx, THREAD_DISPLAY, "display");
}
