#include <pthread.h>
#include <errno.h>
#include <libswscale/swscale.h>

#ifdef USE_LITMUS
#include "litmus.h"
#endif

#include "inc/common.h"
#include "inc/queue.h"
#include "inc/video_queue.h"
#include "inc/thread_ctx.h"
#include "inc/sem_common.h"
#include "inc/profile.h"
#include "inc/video_common.h"
#include "inc/options.h"
#include "inc/timing.h"
#include "inc/litmus_video.h"

static AVFrame* frame;
static uint64_t global_vid_pts = AV_NOPTS_VALUE;

static int get_avframe_buffer(struct AVCodecContext* c, AVFrame* frame)
{
	int ret = avcodec_default_get_buffer(c, frame);
	uint64_t* pts = av_malloc(sizeof(*pts));
	if (!pts)
		die("could not allocate pts\n");
	*pts = global_vid_pts;
	frame->opaque = pts;
	return ret;
}

static void release_avframe_buffer(struct AVCodecContext* c, AVFrame* frame)
{
	if (frame)
		av_freep(&frame->opaque);
	avcodec_default_release_buffer(c, frame);
}

static void set_period(struct thread_ctx* t_ctx)
{
	struct video_state *vs = t_ctx->vs;
	AVRational av_period;
	AVRational fps = vs->fmt_ctx->streams[vs->video_stream]->avg_frame_rate;
	double period;
	char *user_spec;

	if (t_ctx->opts->set_period > 0.0)
	{
		/* user specified period */
		t_ctx->period = t_ctx->opts->set_period * 1e6;
		user_spec = "(user specified) ";
	}
	else
	{

		if (fps.num == 0 || fps.den == 0)
			die("This video does not have an average FPS.\n");

		av_period = av_d2q(1000000000, 1000000000);
		av_period = av_div_q(av_period, fps); /* ns per frame */
		period = av_q2d(av_period);
		t_ctx->period = period;
		user_spec = "\0";
	}

	printf("The %speriod is: %llu nanoseconds\n", user_spec,
			t_ctx->period);
}

static void do_user_seek(struct thread_ctx *t_ctx)
{
	struct video_state *vs = t_ctx->vs;
	struct options *opts = t_ctx->opts;
	const int v_stream = vs->video_stream;
	const AVRational time_base = vs->fmt_ctx->streams[v_stream]->time_base;
	int64_t timestamp;
	int err;

	/* does a * b / c  note inverted fraction b/c we want to divide */
	timestamp = av_rescale(opts->seek_to, time_base.den, time_base.num);
	timestamp /= 1000;

	err = av_seek_frame(vs->fmt_ctx, vs->video_stream, timestamp, 0);
	if (err < 0)
		die("Could not seek to user-specified position");
}

/*
 * This initialize method sets up the global data structures used by all the
 * libav libraries. If anything fails, it exits the program.
 */
void initialize_av(struct thread_ctx *t_ctx)
{
	struct video_state *vs = t_ctx->vs;
	int i, ret;

	/* initialize the video state struct (except fname) */
	vs->video_stream = -1;
	vs->frames_decoded = 0;
	vs->fmt_ctx = NULL;
	vs->codec_ctx = NULL;
	vs->codec = NULL;

	av_register_all();

	if (av_open_input_file(&vs->fmt_ctx, vs->fname, NULL, 0, NULL))
		die("Could not open file\n");

	/* This function populates the streams field */
	if ((ret = av_find_stream_info(vs->fmt_ctx)) < 0)
		die("Could not find stream information\n");

	dump_format(vs->fmt_ctx, 0, vs->fname, 0);

	for (i = 0; i < vs->fmt_ctx->nb_streams; i++) {
		/*
		 * This loop finds which stream has video.
		 */
		vs->codec_ctx = vs->fmt_ctx->streams[i]->codec;

		if (vs->codec_ctx->codec_type == CODEC_TYPE_VIDEO) {
			vs->video_stream = i;
			break;
		}
	}

	if (vs->video_stream == -1)
		die("Could not find video stream\n");

	vs->codec = avcodec_find_decoder(vs->codec_ctx->codec_id);
	if (!vs->codec)
		die("Unsupported codec\n");

	if (avcodec_open(vs->codec_ctx, vs->codec) < 0)
		die("Could not open codec\n");

	vs->codec_ctx->get_buffer = get_avframe_buffer;
	vs->codec_ctx->release_buffer = release_avframe_buffer;

	frame = avcodec_alloc_frame();
	if (!frame)
		die("Couldn't allocate frame\n");

	set_period(t_ctx);
	do_user_seek(t_ctx);
}

static void scale_picture(struct thread_ctx *thread_ctx, AVPicture *dst)
{
	static struct SwsContext *img_convert_ctx = NULL;
	struct video_state *vs = thread_ctx->vs;
	const int width = vs->codec_ctx->width;
	const int height = vs->codec_ctx->height;
	const uint8_t *myData[4];
	int dummy_ret;

	if (!img_convert_ctx)
	{
		img_convert_ctx = sws_getContext(width, height,
				vs->codec_ctx->pix_fmt, width, height,
				SCREEN_PIX_FMT, SWS_BICUBIC, NULL, NULL,
				NULL);
		if (!img_convert_ctx)
			die("Cannot initialize conversion context\n");
	}

	/* The below is because GCC wants const pointers. */
	/* FIXME There should be a way to cast this? */
	myData[0] = frame->data[0]; myData[1] = frame->data[1];
	myData[2] = frame->data[2]; myData[3] = frame->data[3];

	PROFILE_FUNCTION_CALL(thread_ctx, THREAD_DECODE, PROF_SWS_SCALE,
			dummy_ret, sws_scale, img_convert_ctx, myData,
			frame->linesize, 0, height, dst->data, dst->linesize);
}

static inline void enqueue_picture(struct thread_ctx *t_ctx, double pts,
		const int no_block)
{
	struct video_state *vs = t_ctx->vs;
	struct vid_q_item* vq_item;
	char* sem_err_msg;

	if (no_block)
	{
		/* only used when preloading the queue */
		errno = 0;
		if (-1 == sem_trywait(t_ctx->sems->vidq_full))
		{
			if (errno == EAGAIN)
				sem_err_msg = "preload: queue not large enough";
			else
				sem_err_msg = "preload: semaphore error";
			die(sem_err_msg);
		}
	} else
		sem_wait_wrap(t_ctx->sems->vidq_full, t_ctx, THREAD_DECODE,
				vs->frames_decoded);

	/* the queue will point us at a vid_q_item within its memory space */
	if (!queue_start_write(vs->vid_q, (void*) &vq_item)) {
		if (!should_quit())
		{
			/* semaphore should prevent this! */
			die("ERROR: Video queue full! DO NOT IGNORE THIS!");
		}
		else
		{
			/* we should quit, so get out of here */
			return;
		}
	}

	/* scale the picture data from the decoded frame into the queue */
	vq_item->pts = pts;
	scale_picture(t_ctx, &vq_item->picture);

	if (frame->repeat_pict)
		fprintf(stderr, "REPEAT PICT SET!\n");

	queue_finish_write(vs->vid_q, vq_item);
	sem_post_wrap(t_ctx->sems->vidq_empty);
}

/*
 * Returns true if we decoded and enqueued a frame.
 */
static inline int decode_frame(struct thread_ctx *thread_ctx,
		const int no_block)
{
	struct video_state* vs = thread_ctx->vs;
	int frame_finished, ret;
	double pts = 0;

	global_vid_pts = vs->packet.pts;

	PROFILE_FUNCTION_CALL(thread_ctx, THREAD_DECODE, PROF_DECODE_VIDEO, ret,
			avcodec_decode_video2, vs->codec_ctx, frame,
			&frame_finished, &vs->packet);

	if (vs->packet.dts != AV_NOPTS_VALUE && frame->opaque &&
			*(uint64_t*) frame->opaque != AV_NOPTS_VALUE)
	{
		pts = *(uint64_t*)frame->opaque;
	} else if (vs->packet.dts != AV_NOPTS_VALUE) {
		pts = vs->packet.dts;
	}
	pts *= av_q2d(vs->fmt_ctx->streams[vs->video_stream]->time_base);

	if (ret <= 0)
	{
		/*
		 * This doesn't happen ever?
		 */
		die("Could not decompress frame\n");
	}

	if (frame_finished)
	{
		++(vs->frames_decoded);
		enqueue_picture(thread_ctx, pts, no_block);
	} else {
		/*
		 * Happens only on frame one?
		 */
		fprintf(stderr, "Frame finished returned zero!\n");
	}

	return frame_finished;
}

/*
 * Decode one video frame and place it in the queue, unless we are at the end of
 * the video (or there is an error), in which case we return non-zero.
 */
static int decode_video_frame(struct thread_ctx* thread_ctx, const int no_block)
{
	struct video_state* vs = thread_ctx->vs;
	int got_video_frame = 0, decoded_frame = 0, err = 0;

	do {
		/*
		 * On av_read_frame: Must free with av_free_packet() For video,
		 * the packet contains exactly one frame
		 */
		PROFILE_FUNCTION_CALL(thread_ctx, THREAD_DECODE, PROF_READ_FRAME,
				err, av_read_frame, vs->fmt_ctx, &vs->packet);

		if (!err)
		{
			if (vs->packet.stream_index == vs->video_stream) 
			{
				decoded_frame = decode_frame(thread_ctx,
						no_block);

				if (decoded_frame) {
					got_video_frame = 1;
				}
			}

			/*
			 * Must free memory inside the packet that was
			 * allocated by av_read_frame().
			 */
			av_free_packet(&vs->packet);
		}
		/* Stop if there is an error reading or we get a video frame */
	} while (!err && !got_video_frame);

	return err;
}

/*
 * Called when the pthread exits, so we can get the total CPU time taken.
 */
static void thread_cleanup_prof_completion(void* thread_ctx_in)
{
	struct thread_ctx* thread_ctx = (struct thread_ctx*) thread_ctx_in;
	unsigned long long total_time = cpu_time();
	prof_completion_event(thread_ctx, THREAD_DECODE, total_time,
			thread_ctx->vs->frames_decoded);
}

static void preload_queue(struct thread_ctx *t_ctx)
{
	const int num_to_preload = frames_to_preload(t_ctx);
	int i, err;

	for (i = 0; i < num_to_preload; ++i)
	{
		err = decode_video_frame(t_ctx, 1);
		if (err)
			die("Could not preload queue.");
	}

	signal_preload_queue();
}

/*
 * Return true if we've decoded enough frames to fulfill the user-specified
 * quitting time.
 */
static inline int time_is_up(struct thread_ctx *t_ctx)
{
	static unsigned long long stop_frame = 0;
	static unsigned long long stop_ns;

	if (!stop_frame)
	{
		/* initialize */
		if (t_ctx->opts->stop_after != 0.0)
		{
			stop_ns = t_ctx->opts->stop_after * 1e6;
			stop_frame = stop_ns / t_ctx->period;
			if (stop_frame * t_ctx->period != stop_ns)
				stop_frame++;
		}
		else
			stop_frame = ULLONG_MAX;
	}

	return (stop_frame <= t_ctx->vs->frames_decoded);
}

void* decode_thread_loop(void* thread_ctx_in)
{
	struct thread_ctx* thread_ctx = (struct thread_ctx*) thread_ctx_in;
	int err;
	int *ret = NULL;
#ifdef USE_PROFILE
	unsigned long long start, end;
#endif

	ret = av_malloc(sizeof(*ret));
	if (!ret)
		die("Could not allocate decode thread return value");
	*ret = 1;

	pthread_cleanup_push(thread_cleanup_prof_completion, thread_ctx);

	preload_queue(thread_ctx);

	/* not used if non-LITMUS^RT */
	make_litmus_decoder(thread_ctx);
	wait_for_release();

	do {

#ifdef USE_PROFILE
		start = cpu_time();
#endif

		err = decode_video_frame(thread_ctx, 0);

		if (time_is_up(thread_ctx))
			err = 1;

#ifdef USE_PROFILE
		end = cpu_time();
		prof_start_end(thread_ctx, THREAD_DECODE,
				PROF_FULL_DECODE, start, end);
#endif

		if (!err)
			litmus_job_completion();
	} while (!err);

	thread_ctx->decoder_done = 1;
	/* Free a stuck display thread from a possible race condition (see
	 * comments in other file) */
	sem_post_wrap(thread_ctx->sems->vidq_empty);
	pthread_cleanup_pop(1);
	become_regular_task();
	pthread_exit(ret);
}

void tear_down_av(struct video_state *vs)
{
	av_free(frame);

	if (avcodec_close(vs->codec_ctx))
		die("avcodec_close() failed?!\n");

	av_close_input_file(vs->fmt_ctx);
}
