#include "inc/common.h"
#include "inc/queue.h"
#include "inc/video_common.h"
#include "inc/options.h"
#include "inc/thread_ctx.h"
#include "inc/video_queue.h"

/* This is used so we can deallocate any AVPicture objects at exit. */
static struct avpicture_node* pict_head = NULL;

static inline void vid_q_avpicture_swap(AVPicture *a, AVPicture *b)
{
	const AVPicture t = *a;
	*a = *b;
	*b = t;
}

static void vid_q_item_swap(struct vid_q_item *dst, struct vid_q_item *src)
{
	vid_q_avpicture_swap(&dst->picture, &src->picture);
	dst->pts = src->pts;
}

/* how large we will allow the queue to be 2^(POW_TWO_MAX - 1) */
#define POW_TWO_MAX	6
static int next_pow_two(const int test)
{
	static int pow_two[POW_TWO_MAX] = { 0 };
	int i;

	if (pow_two[0] == 0)
	{
		for (i = 0; i < POW_TWO_MAX; ++i)
			pow_two[i] = (1 << i);
	}

	for (i = 0; i < POW_TWO_MAX; ++i)
	{
		if (test <= pow_two[i])
			break;
	}

	if (i == POW_TWO_MAX)
		die("Your queue size is too large.");

	return pow_two[i];
}

/* how many video frames we should preload */
int frames_to_preload(struct thread_ctx *t_ctx)
{
	/* convert milliseconds to nanoseconds */
	const unsigned long long time = t_ctx->opts->preload_time * 1e6;
	int frames = time / t_ctx->period;

	if (time > 0 && time % t_ctx->period > 0)
		frames++;

	return frames;
}

/* defined below */
static void save_avpicture(AVPicture* pict);

void initialize_vid_q(struct thread_ctx *t_ctx)
{
	struct video_state *vs = t_ctx->vs;
	unsigned int slot_count = next_pow_two(frames_to_preload(t_ctx));
	struct vid_q_item* vid_q_items = NULL, *vq_item;
	const int width = vs->codec_ctx->width;
	const int height = vs->codec_ctx->height;
	AVPicture *picture;
	char* slots = NULL;
	int err;

	if (slot_count < 1)
		die("The queue size was zero!\n");

	t_ctx->queue_size = slot_count;

	vid_q_items = av_malloc(sizeof(*vid_q_items) * slot_count);
	if (!vid_q_items)
		die("Could not allocate video queue\n");

	slots = av_malloc(slot_count);
	if (!slots)
		die("Could not allocate slot indicator space\n");

	for (vq_item = vid_q_items; vq_item < &vid_q_items[slot_count];
			++vq_item)
	{
		/* allocate the AVPicture data structures now, saving for
		 * later deallocation */
		picture = &(vq_item->picture);
		avpicture_alloc_wrap(picture, width, height);
		save_avpicture(picture);
	}

	err = queue_init(&vs->vid_q, slot_count, sizeof(*vid_q_items), slots,
			vid_q_items,
			(void(*)(void*, void*))vid_q_item_swap);
	if (!err)
		die("could not initialize queue\n");
}

/* defined below */
static void destroy_avpictures(void);

void tear_down_vid_q(struct video_state* vs)
{
	destroy_avpictures();
	queue_destroy(&vs->vid_q);
}


/*********************************************************************
 * AVPicture list stuff
*********************************************************************/

struct avpicture_node {
	AVPicture* picture;
	struct avpicture_node* next;
};

static struct avpicture_node* new_avpicture_node(void)
{
	struct avpicture_node* node;
	node = av_malloc(sizeof(*node));
	if (!node)
		die("Could not allocate avpicture node\n");
	node->picture = NULL;
	node->next = NULL;
	return node;
}

static void save_avpicture(AVPicture* pict)
{
	struct avpicture_node* node = new_avpicture_node();
	node->picture = pict;
	node->next = pict_head;
	pict_head = node;
}

static void destroy_avpictures(void)
{
	struct avpicture_node* tmp, *node = pict_head;

	while (node != NULL) {
		tmp = node;
		node = node->next;
		pict_head = node;
		avpicture_free(tmp->picture);
		av_free(tmp);
	}
}
